﻿using FLY.Modbus;
using FLY.OBJComponents.IService;
using FLY.OBJComponents.Server;
using FLY.Weight.Common;
using FLY.Weight.IService;
using FLY.Weight.Server.Model;
using FObjBase;
using Newtonsoft.Json;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Text.RegularExpressions;

namespace FLY.Weight.Server
{
    [JsonObject(MemberSerialization.OptIn)]
    public class WeightSystem : IWeightSystemService, IPropertyOpt
    {
        const int FlowIntervalSec = 10;
        #region IWeightSystemService 接口

        public ObservableCollection<WeighterC> Items { get; } = new ObservableCollection<WeighterC>();

        /// <summary>
        /// 每1秒保存一次配料数据， 一共10秒。
        /// 当卸料时，提起3秒前的数据保存
        /// </summary>
        List<Dictionary<WeighterC, Lc_Mix>> mix10 = new List<Dictionary<WeighterC, Lc_Mix>>();

        public WeighterAccessory Accessory { get; } = new WeighterAccessory();

        private PLCProxySystem plcos = new PLCProxySystem();
        /// <summary>
        /// PLC代理系统
        /// </summary>
        public IPLCProxySystemService PLCos { get { return plcos; } }


        /// <summary>
        /// 层数
        /// </summary>
        public int ItemsCnt { get; private set; } = 3;

        /// <summary>
        /// 每层仓数
        /// </summary>
        public int[] BinCnts { get; private set; } = new int[] { 4, 4, 4 };

        /// <summary>
        /// 螺杆频率调节增量， 用于 +,- 号
        /// </summary>
        public double ScrewMotorFreqInc { get; set; } = 0.1;

        /// <summary>
        /// 使用的界面版本信息
        /// </summary>
        public string Version { get; set; } = "version=3;itemStyle=0";

        /// <summary>
        /// 每层 的名称，null 就是使用默认的 ABC
        /// </summary>
        public string[] NumberNames { get; set; }

        /// <summary>
        /// plcgroup.json 文件版本信息
        /// </summary>
        public string FileVersion { get; set; }
        #endregion


        /// <summary>
        /// 报警系统
        /// </summary>
        WarningSystem2 warning;
        /// <summary>
        /// 记录到数据库
        /// </summary>
        HistoryDb historyDb;
        /// <summary>
        /// 报警配置
        /// </summary>
        ErrorConf errorConf;
        /// <summary>
        /// 周期保存Ibc数据
        /// </summary>
        PeriodicallySaveData<Lc_Flow> psdFlow;
        public WeightSystem()
        {
            LoadNumberNames();
            Load();
            //--------------------------------------------------------------------------------
            //step 1 加载PLC寄存器 文件
            AddConfigFile();

            
        }

        public void Init(HistoryDb historyDb, WarningSystem2 warning)
        {
            this.historyDb = historyDb;
            this.warning = warning;

            //--------------------------------------------------------------------------------
            //step 2 报警配置
            errorConf = new ErrorConf(plcos, this.warning, "称重");
            errorConf.AddErrorAction(Accessory);
            for (int i = 0; i < Items.Count(); i++)
            {
                errorConf.AddErrorAction(
                    Items[i],
                        (ref string description, object state) => {
                            int idx = (int)state;
                            description = Items[idx].Number + "层 " + description;
                        },
                        i);
            }
            errorConf.InitError();



            Misc.BindingOperations.SetBinding(Items[0], nameof(WeighterC.AlarmIsOn), () =>
            {
                this.warning.Enable = !Items[0].AlarmIsOn;
            });
            Misc.BindingOperations.SetBinding(this.warning, nameof(this.warning.Enable), () =>
            {
                for (int i = 0; i < ItemsCnt; i++)
                {
                    Items[i].AlarmIsOn = !this.warning.Enable;
                }
            });
            Misc.BindingOperations.SetBinding(this.warning, nameof(this.warning.IsRinging), () =>
            {
                if (!this.warning.IsRinging)
                {
                    for (int i = 0; i < ItemsCnt; i++)
                    {
                        Items[i].IsAlarmReseted = true;
                        Items[i].IsAlarmReseted = false;
                    }
                }
            });

            //--------------------------------------------------------------------------------
            //step 3 历史数据记录

            //--------------------------------------------------------------------------------
            //配料记录
            foreach (WeighterC w in Items)
            {
                //LastCumulativeProductions.Add(w, w.CumulativeProduction);
                w.PropertyChanged += WeightSystem_MBServer_PropertyChanged;
            }
            //--------------------------------------------------------------------------------
            //Test();


            

            //--------------------------------------------------------------------------------
            //添加任务
            for (int i = 0; i < Items.Count(); i++)
            {
                List<string> props = new List<string>
                {
                    nameof(WeighterC.BucketValveIsOpen),
                    nameof(WeighterC.IsMixSave),
                    nameof(WeighterC.MixDisp),
                    nameof(WeighterC.ClearProduction),
                    nameof(WeighterC.CumulativeProduction),
                    nameof(WeighterC.CurrentFlow),
                    nameof(WeighterC.ScrewPSet),
                    nameof(WeighterC.ScrewPDisp),
                    nameof(WeighterC.ScrewMotorFreq),

                    //报警复位用
                    nameof(WeighterC.IsAlarmReseted),
                    nameof(WeighterC.AlarmIsOn)
                };

                for (int j = 0; j < Items[i].BinCnt; j++)
                {
                    int no = j + 1;
                    string nostr = $"_{no}";
                    props.Add(nameof(IngredientCell.MixSet) + nostr);
                    props.Add(nameof(IngredientCell.MixPSet) + nostr);
                    props.Add(nameof(IngredientCell.MixDisp) + nostr);
                    props.Add(nameof(IngredientCell.MixPDisp) + nostr);
                }

                PLCos.SetPlan($"{nameof(Items)}[{i}]", props, 0);
            }
            PLCos.SetPlan(nameof(Accessory), new string[] { nameof(Accessory.TotalFlow) }, 0);


            //提前记录配料数据到内存。 重新一次配料, 把之前保存的内存的数据，保存到数据库。
            FObjBase.PollModule.Current.Poll_Config(PollModule.POLL_CONFIG.ADD, OnPoll_SaveMix, TimeSpan.FromSeconds(1));

            //周期保存每层流量数据
            psdFlow = new PeriodicallySaveData<Lc_Flow>();
            psdFlow.Init(
                FlowIntervalSec,
                this.historyDb.AddFlowRange,
                () => {
                    if (!PLCos.IsConnectedWithPLC)
                        return null;
                    
                    if (Accessory.TotalFlow < 5)//没有生产
                        return null;

                    //记录数据
                    Lc_Flow lc_Flow = new Lc_Flow
                    {
                        Time = DateTime.Now,
                        Total = Math.Round(Accessory.TotalFlow, 4),
                        Items = new FlyData_FlowItem[Items.Count]
                    };
                    for (int i = 0; i < Items.Count; i++)
                    {
                        lc_Flow.Items[i] = new FlyData_FlowItem
                        {
                            Flow = Math.Round(Items[i].CurrentFlow, 4),
                            ScrewPDisp = Math.Round(Items[i].ScrewPDisp, 2),
                            ScrewMotorFreq = Math.Round(Items[i].ScrewMotorFreq, 1)
                        };
                    }
                    return lc_Flow;
                });

            //--------------------------------------------------------------------------------
            //last step  PLC 更新计划服务初始化
            plcos.Init();


            this.PropertyChanged += WeightSystem_PropertyChanged;
        }

        private void WeightSystem_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            if (e.PropertyName == nameof(ScrewMotorFreqInc)) {
                Save();
            }
        }

        void OnPoll_SaveMix()
        {
            Dictionary<WeighterC, Lc_Mix> mix = new Dictionary<WeighterC, Lc_Mix>();
            foreach (WeighterC w in Items)
            {
                //记录
                Lc_Mix f = new Lc_Mix
                {
                    Time = DateTime.Now,
                    Total = w.MixDisp
                };
                f.Items = new FlyData_MixItem[w.Ingredients.Count];
                for (int i = 0; i < w.Ingredients.Count; i++)
                {
                    FlyData_MixItem item = new FlyData_MixItem()
                    {
                        Disp = w.Ingredients[i].MixDisp,
                        PDisp = w.Ingredients[i].MixPDisp
                    };
                    f.Items[i] = item;
                }
                mix.Add(w, f);
            }
            mix10.Add(mix);
            if (mix10.Count > 10)
                mix10.RemoveAt(0);
        }


        //每隔1秒记录，w.MixDisp, Ingredients[i].(MixDisp,MixPDisp)，保存10s
        //当卸料时，提取2s前的数据保存！！！！
        private void WeightSystem_MBServer_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            WeighterC w = sender as WeighterC;

            if (e.PropertyName == nameof(w.IsMixSave))//if (e.PropertyName == "BucketValveIsOpen")
            {
                if (w.IsMixSave)// (w.BucketValveIsOpen)
                {
                    Lc_Mix f_last = null;
                    //提取之前的记录
                    if (mix10.Count > 3)
                    {
                        f_last = mix10[mix10.Count - 3][w];
                    }


                    //记录
                    Lc_Mix f = new Lc_Mix()
                    {
                        Time = DateTime.Now,
                        Items = new FlyData_MixItem[w.Ingredients.Count]
                    };

                    if (f_last != null)// && f_last.Total > f.Total)
                    {
                        //没错,用以前的是对的！！！！
                        f.Total = Math.Round(f_last.Total, 4);
                        for (int i = 0; i < w.Ingredients.Count; i++)
                        {
                            FlyData_MixItem item = new FlyData_MixItem()
                            {
                                Disp = Math.Round(f_last.Items[i].Disp, 4),
                                PDisp = Math.Round(f_last.Items[i].PDisp, 2)
                            };
                            f.Items[i] = item;
                        }
                    }
                    else
                    {
                        f.Total = Math.Round(w.MixDisp, 4);
                        for (int i = 0; i < w.Ingredients.Count; i++)
                        {
                            FlyData_MixItem item = new FlyData_MixItem()
                            {
                                Disp = Math.Round(w.Ingredients[i].MixDisp, 4),
                                PDisp = Math.Round(w.Ingredients[i].MixPDisp, 2)
                            };
                            f.Items[i] = item;
                        }
                    }

                    int index = Items.IndexOf(w);
                    historyDb.AddMix(index, f);
                }
            }
            else if (e.PropertyName == nameof(w.ClearProduction))
            {
                //当累计产量变小，肯定是被clear() 需要保存数据
                if (w.ClearProduction)
                {
                    Lc_Mix f = new Lc_Mix()
                    {
                        Time = DateTime.Now,
                        Items = new FlyData_MixItem[w.Ingredients.Count]
                    };

                    f.Total = Math.Round(w.CumulativeProduction, 4);
                    for (int i = 0; i < w.Ingredients.Count; i++)
                    {
                        FlyData_MixItem item = new FlyData_MixItem()
                        {
                            Disp = Math.Round(w.Ingredients[i].MixCum, 4),
                            PDisp = Math.Round(w.Ingredients[i].MixCumPercent, 2)
                        };
                        f.Items[i] = item;
                    }
                    int index = Items.IndexOf(w);
                    historyDb.AddRollMix(index, f);
                }
                //LastCumulativeProductions[w] = w.CumulativeProduction;
            }
        }

        void Test()
        {
            #region 测试
            Items[0].CurrentFlow = 150;
            Items[0].BinWeight = 15;
            Items[0].MixBucketWeight = 4;
            Items[0].MixPSet_1 = 90;
            Items[0].MixPSet_2 = 5;
            Items[0].MixPSet_3 = 5;
            Items[0].MixPSet_4 = 0;

            Items[1].CurrentFlow = 100;
            Items[1].BinWeight = 10;
            Items[1].MixBucketWeight = 4;
            Items[1].MixPSet_1 = 20;
            Items[1].MixPSet_2 = 70;
            Items[1].MixPSet_3 = 5;
            Items[1].MixPSet_4 = 5;

            Items[2].CurrentFlow = 50;
            Items[2].BinWeight = 5;
            Items[2].MixBucketWeight = 4;
            Items[2].MixPSet_1 = 0;
            Items[2].MixPSet_2 = 70;
            Items[2].MixPSet_3 = 30;
            Items[2].MixPSet_4 = 0;

            Items[2].IsErrorOfAdd_1 = true;
            Items[0].IsErrorOfBlender = true;

            Accessory.TotalFlow = Items.Sum((w) => { return w.CurrentFlow; });

            foreach (WeighterC w in Items)
            {
                w.FlowSetting = w.CurrentFlow;
                w.ScrewPDisp = w.CurrentFlow * 100 / Accessory.TotalFlow;
                w.ScrewPSet = w.ScrewPDisp;
                w.ScrewMotorFreq = 100 * w.CurrentFlow / 300;
                w.CumulativeProduction = w.CurrentFlow * 20;

                w.MixPDisp_1 = w.MixPSet_1;
                w.MixPDisp_2 = w.MixPSet_2;
                w.MixPDisp_3 = w.MixPSet_3;
                w.MixPDisp_4 = w.MixPSet_4;

                w.MixCum_1 = (float)(w.CumulativeProduction * w.MixPSet_1 / 100.0);
                w.MixCum_2 = (float)(w.CumulativeProduction * w.MixPSet_2 / 100.0);
                w.MixCum_3 = (float)(w.CumulativeProduction * w.MixPSet_3 / 100.0);
                w.MixCum_4 = (float)(w.CumulativeProduction * w.MixPSet_4 / 100.0);

                w.MixCumPercent_1 = w.MixPSet_1;
                w.MixCumPercent_2 = w.MixPSet_2;
                w.MixCumPercent_3 = w.MixPSet_3;
                w.MixCumPercent_4 = w.MixPSet_4;
            }


            Accessory.TotalProduction = (UInt16)Items.Sum((w) => { return w.CumulativeProduction; });
            Accessory.TotalFlowSetting = Accessory.TotalFlow;


            Random r = new Random();
            Lc_Flow last_f = null;
            int flowListSize = 10000;
            DateTime startTime = DateTime.Now - TimeSpan.FromSeconds(flowListSize * FlowIntervalSec);
            for (int j = 0; j < flowListSize; j++)
            {
                Lc_Flow f = new Lc_Flow();
                f.Time = startTime + TimeSpan.FromSeconds(FlowIntervalSec);

                f.Items = new FlyData_FlowItem[ItemsCnt];
                for (int i = 0; i < Items.Count; i++)
                {
                    f.Items[i] = new FlyData_FlowItem();
                    f.Items[i].Flow = (float)(Items[i].CurrentFlow + r.NextDouble() * 5 - 2.5);
                    f.Items[i].ScrewPDisp = Items[i].ScrewPDisp;
                    f.Items[i].ScrewMotorFreq = Items[i].ScrewMotorFreq;

                    //f.Items[i].CumulativeProduction = (float)(f.Items[i].Flow * FlowInterval / 60);
                    //if (last_f != null)
                    //{
                    //    f.Items[i].CumulativeProduction += last_f.Items[i].CumulativeProduction;
                    //}

                    //f.Items[i].MixCum[0] = (float)(f.Items[i].CumulativeProduction * Items[i].MixPSet_1 / 100.0);
                    //f.Items[i].MixCum[1] = (float)(f.Items[i].CumulativeProduction * Items[i].MixPSet_2 / 100.0);
                    //f.Items[i].MixCum[2] = (float)(f.Items[i].CumulativeProduction * Items[i].MixPSet_3 / 100.0);
                    //f.Items[i].MixCum[3] = (float)(f.Items[i].CumulativeProduction * Items[i].MixPSet_4 / 100.0);


                    //f.Items[i].MixCumPercent[0] = Items[i].MixPSet_1;
                    //f.Items[i].MixCumPercent[1] = Items[i].MixPSet_2;
                    //f.Items[i].MixCumPercent[2] = Items[i].MixPSet_3;
                    //f.Items[i].MixCumPercent[3] = Items[i].MixPSet_4;
                }
                f.Total = f.Items.Sum(i => i.Flow);
                last_f = f;
                historyDb.AddFlow(f);
            }
            #endregion
        }


        string GetNumber(int index)
        {
            string number;

            if (NumberNames != null
                && index < NumberNames.Count()
                && !string.IsNullOrEmpty(NumberNames[index]))
            {
                number = NumberNames[index];
            }
            else
            {
                number = ((char)('A' + index)).ToString();
            }
            return number;
        }
        void AddConfigFile()
        {
            plcos.AddConfigFile("plcgroup.json", (plcgroup) =>
            {
                //从plcgroup 分析出 有多少层，每层的仓数
                List<int> bincnts = new List<int>();
                Regex r = new Regex(@"Items\[([0-9])\]");
                Regex r2 = new Regex("MixPSet_([1-9])");

                foreach (PLCGroup.PLCVariable var in plcgroup.Variables)
                {
                    Match m = r.Match(var.OwnerName);
                    if (m.Success)
                    {
                        int weighter_idx = int.Parse(m.Groups[1].Value);
                        if (bincnts.Count() <= weighter_idx)
                        {
                            while (bincnts.Count() <= weighter_idx)
                                bincnts.Add(0);
                        }
                        Match m2 = r2.Match(var.PropertyName);
                        if (m2.Success)
                        {
                            int bincnt = int.Parse(m2.Groups[1].Value);
                            if (bincnts[weighter_idx] < bincnt)
                                bincnts[weighter_idx] = bincnt;
                        }
                    }
                }
                BinCnts = bincnts.ToArray();
                ItemsCnt = BinCnts.Count();

                for (int i = 0; i < BinCnts.Count(); i++)
                {
                    string number = GetNumber(i);
                    WeighterC w = new WeighterC(number, i, BinCnts[i]);
                    Items.Add(w);
                }


                //objname 转 obj
                Dictionary<string, INotifyPropertyChanged> objnames = new Dictionary<string, INotifyPropertyChanged>();

                objnames.Add(nameof(Accessory), Accessory);
                for (int i = 0; i < ItemsCnt; i++)
                    objnames.Add($"{nameof(Items)}[{i}]", Items[i]);
                return objnames;
            });

            //这个是从测厚仪获取的，每几秒就写一次，不调试！！！！！
            plcos.IgnoreLogProperties.Add(new SenderProperty() { sender = Accessory, propertyName = nameof(Accessory.CurrentVelocitySet) });

        }

        public event PropertyChangedEventHandler PropertyChanged;


        string filePath_number = "weight2NumberNames.json";
        bool LoadNumberNames()
        {
            if (!File.Exists(filePath_number))
                return false;
            string json = File.ReadAllText(filePath_number);
            var p = Newtonsoft.Json.JsonConvert.DeserializeObject<NumberNamesJsonDb>(json);
            NumberNames = p.NumberNames;
            Version = p.VersionString;
            FileVersion = p.FileVersion;
            return true;
        }
        public void SaveNumberNames()
        {
            NumberNamesJsonDb jsonDb = new NumberNamesJsonDb
            {
                NumberNames = NumberNames,
                VersionString = Version,
                FileVersion = FileVersion
            };
            string json = Newtonsoft.Json.JsonConvert.SerializeObject(jsonDb);
            File.WriteAllText(filePath_number, json);
        }

        string filePath = "weightsystem.json";
        bool Load() {
            return WeightSystemJsonDb.Load(filePath, this);
        }
        bool Save()
        {
            return WeightSystemJsonDb.Save(filePath, this);
        }


        public string[] GetSyncPropNames()
        {
            return new string[]{
                nameof(ItemsCnt),
                nameof(BinCnts),
                nameof(NumberNames),
                nameof(Version),
                nameof(ScrewMotorFreqInc)
                };
        }

        public string[] GetNoSyncPropNames()
        {
            return null;
        }
    }
    public class NumberNamesJsonDb
    {
        /// <summary>
        /// 每层 的名称，null 就是使用默认的 ABC
        /// </summary>
        public string[] NumberNames;
        /// <summary>
        /// 使用的界面版本信息
        /// </summary>
        public string VersionString;
        /// <summary>
        /// plcgroup.json 文件版本信息
        /// </summary>
        public string FileVersion;
    }
    public class WeightSystemJsonDb 
    {
        public static bool Save(string filePath, WeightSystem src) 
        {
            try
            {
                WeightSystemJsonDb jsonDb = new WeightSystemJsonDb()
                {
                    ScrewMotorFreqInc = src.ScrewMotorFreqInc
                };
                string json = Newtonsoft.Json.JsonConvert.SerializeObject(jsonDb);
                File.WriteAllText(filePath, json);
                return true;
            }
            catch {
                return false;
            }
        }
        public static bool Load(string filePath, WeightSystem src)
        {
            if (!File.Exists(filePath)) 
            {
                return false;
            }
            try
            {
                string json = File.ReadAllText(filePath);
                var jsonDb = Newtonsoft.Json.JsonConvert.DeserializeObject<WeightSystemJsonDb>(json);
                src.ScrewMotorFreqInc = jsonDb.ScrewMotorFreqInc;
                return true;
            }
            catch {
                return false;
            }
        }
        /// <summary>
        /// 螺杆频率调节增量， 用于 +,- 号
        /// </summary>
        public double ScrewMotorFreqInc = 0.1;
    }
}
